using NUnit.Framework;
using System.Linq;

using Encompass;
using System.Collections.Generic;
using System;

namespace Tests {
    public class EngineTest
    {
        static IEnumerable<MockComponent> components;
        static MockComponent component;

        public class ReadComponentsTestEngine : Engine
        {
            public override void Update(float dt)
            {
                components = this.ReadComponents<MockComponent>();
            }
        }

        public class ReadComponentTestEngine : Engine
        {
            public override void Update(float dt)
            {
                component = this.ReadComponent<MockComponent>();
            }
        }

        [Test]
        public void ReadComponents()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine<ReadComponentsTestEngine>();

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            MockComponent mockComponentB;
            mockComponentB.myInt = 1;
            mockComponentB.myString = "howdy";

            entity.AddComponent(mockComponent);
            entity.AddComponent(mockComponentB);

            var world = worldBuilder.Build();

            world.Update(0.01f);

            Assert.Contains(mockComponent, components.ToList());
            Assert.Contains(mockComponentB, components.ToList());
        }

        [Test]
        public void ReadComponent()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine<ReadComponentTestEngine>();

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            entity.AddComponent(mockComponent);

            var world = worldBuilder.Build();

            world.Update(0.01f);

            Assert.AreEqual(mockComponent, component);
        }

        [Test]
        public void ReadComponentWhenMultipleComponents()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine<ReadComponentTestEngine>();

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            MockComponent mockComponentB;
            mockComponentB.myInt = 1;
            mockComponentB.myString = "howdy";

            entity.AddComponent(mockComponent);
            entity.AddComponent(mockComponentB);

            var world = worldBuilder.Build();

            Assert.Throws<InvalidOperationException>(() => world.Update(0.01f));
        }

        [Mutates(typeof(MockComponent))]
        public class UpdateComponentTestEngine : Engine
        {
            public override void Update(float dt)
            {
                var originalComponent = this.ReadComponent<MockComponent>();
                var newComponent = this.ReadComponent<MockComponent>();
                newComponent.myInt = 420;
                newComponent.myString = "blaze it";
                this.UpdateComponent(originalComponent, newComponent);
                component = this.ReadComponent<MockComponent>();
            }
        }

        [Test]
        public void UpdateComponent()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine<UpdateComponentTestEngine>();

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            entity.AddComponent(mockComponent);

            var world = worldBuilder.Build();

            world.Update(0.01f);

            Assert.AreEqual(420, component.myInt);
            Assert.AreEqual("blaze it", component.myString);
        }

        public class UndeclaredUpdateComponentTestEngine : Engine
        {
            public override void Update(float dt)
            {
                var originalComponent = this.ReadComponent<MockComponent>();
                var newComponent = this.ReadComponent<MockComponent>();
                newComponent.myInt = 420;
                newComponent.myString = "blaze it";
                this.UpdateComponent(originalComponent, newComponent);
                component = this.ReadComponent<MockComponent>();
            }
        }

        [Test]
        public void UpdateUndeclaredComponent()
        {
            var worldBuilder = new WorldBuilder();
            worldBuilder.AddEngine<UndeclaredUpdateComponentTestEngine>();

            var entity = worldBuilder.CreateEntity();

            MockComponent mockComponent;
            mockComponent.myInt = 0;
            mockComponent.myString = "hello";

            entity.AddComponent(mockComponent);

            var world = worldBuilder.Build();

            Assert.Throws<IllegalComponentMutationException>(() => world.Update(0.01f));
        }
    }
}
